/* (c) Copyright 1993 by Bruce W. Watson */

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "pattm.h"
#include "funcs.h"



#define PREF (500)
#define MAXSET (100)
#define STRING (1100000)
#define MAXMATCH (500000)
#define KEYWORDTEXT (1000)


static FILE *stringfile, *setfile, *acfailfile, *acoptfile, *cwnormfile, *cwbmfile;


static char S[STRING];
static char *Sprime = S + 1;

/* mainly for KMP */
static int f_hat[PREF];
static occ_kmp_bm_idx_t O_hat[MAXMATCH];

/* mainly for AC */
static pref_t tau[PREF][V_SIZE], gamma_f[PREF][V_SIZE], f[PREF];
static kw_t P[MAXSET], *Output[PREF];

static char p[KEYWORDTEXT];

static occ_idx_t O[MAXMATCH];
/* mainly for CW */
static suff_t tau_r[PREF][V_SIZE];

static int m_P, Shift[PREF][V_SIZE], d1[PREF], d2[PREF],
	Char[V_SIZE], d3_bar[V_SIZE], suff_lengths[PREF];
static suff_t last, scratch[PREF], drop[PREF];


static void printmatches( FILE *outfile, occ_idx_t *Occ, occ_idx_t *E ) {
	auto const kw_t *Plocal;
	
	fputs( "\n", outfile );
	while( Occ != E ) {
		for( Plocal = Occ->what; Plocal != NULL; Plocal = Plocal->output_next ) {
			fprintf( outfile, "%s:%ld\n", Plocal->p, (long)(Occ->where) );
		}
		Occ++;
	}
	return;
}


static long readstring( void ) {
	/* read in the input string, append zeroes on the end */
	auto long i;
	S[0] = '\0';

	for( i = 0; (Sprime[i] = fgetc( stringfile )) != EOF; i++ );
	while( i < STRING ) {
		Sprime[i++] = 0;
	}
	return( i );
}


static void touch( void ) {
	auto long i;
	auto unsigned long k;

	for( i = 0; i < STRING; i++ ) {
		k |= (unsigned)S[i];
	}
	printf( "Touched %ld\n", (long)k );
	return;
}

static int readset( void ) {
	auto int done; /* done kw's have already been read */
	auto int i, begin;
	auto int c;
	auto kw_t *scratch;
	auto int totlen;

	/* read in a set of keywords (on one line), return zero
		if not possible, nonzero if possible,
		print the set */
	while( isspace( c = fgetc( setfile ) ) );
	if( c == EOF ) {
		return( 0 );
	}
	done = i = 0;
	while( (c != '\n') && (c != EOF) ) {
		/* still something to be read on this line */
		begin = i;
		while( isalpha( p[i] = c ) ) {
			i++; c = fgetc( setfile );
		}
		p[i] = '\0';
		P[done].p = &(p[begin]);
		P[done].length = (i - begin);
		i++;
		P[done].next = P+done+1; 
		P[done].output_next = NULL;
		done++;

		while( c == ' ' ) {
			c = fgetc( setfile );
		}
	}
	P[done - 1].next = NULL; /* mark the end */

	printf( "Keywords" );
	for( scratch = P, totlen = 0; !Empty(scratch); scratch = Next(scratch) ) {
		printf( " %s", scratch->p );
		totlen += scratch->length;
	}
	printf( "\nTotallength %d\n", totlen );

	return( done );
}


static void inittimer( void ) {
	auto struct rusage tb, te;
	auto int r1, r2;
	/* call the timer fetch a couple of times */
	/* print the null timer fetch */
	getrusage( RUSAGE_SELF, &tb );
	r1 = getrusage( RUSAGE_SELF, &tb );
	r2 = getrusage( RUSAGE_SELF, &te );
	if( r1 || r2 ) {
		fputs( "getrusage failure", stderr );
		exit( -1 );
	}
	printf( "Nulltimer %ld %ld\n",
		(long)(te.ru_utime.tv_sec - tb.ru_utime.tv_sec),
		(long)(te.ru_utime.tv_usec - tb.ru_utime.tv_usec) );
	return;
}

static void KMP( void ) {
	/* assume that it is a single keyword */
	f_hat_ef_precomp( P->p, P->length, f_hat );
	kmp_idx( Sprime, O_hat, f_hat, P->p, P->length ); 
}

static void ACF( void ) {
	/* assume that nothing is precomputed */
	last = tau_ef_precomp( P, tau, Output );
	printf( "Prefsize %ld\n", (long)(last + 1) );

	/* tau contains tau_ef, Output contains ACOutput */
	/* now precompute gamma_f and f_f */

	gamma_f_precomp( gamma_f, f, tau, scratch );
	/* f contains f_f */

	/* assume that the following is instrumented */
	ac_fail_idx( Sprime, Output, tau, O, f );
}

static void ACO( void ) {
	/* assume that tau contains tau_ef
		f contains f_f
		Output contains ACOutput */
	/* tau contains gamma_f */

	/* assume that the following is instrumented */
	ac_opt_idx( Sprime, Output, gamma_f, O );
}

static void CWBM( void ) {
	/* assume that nothing is precomputed */
	cw_d3_bar_precomp( P, d3_bar );

	m_P = cw_m_P_precomp( P );
	printf( "Minkeyword %d\n", m_P );

	cw_char_precomp( d3_bar, Char, m_P );

	/* first calculate tau for g_r purposes */
	tau_r_precomp( P, tau, Output, suff_lengths );

	last = tau_r_precomp( P, tau_r, Output, suff_lengths );
	/* tau is tau_r, Output is CWOutput,
		suff_lengths and last are correct */
	printf( "Suffsize %ld\n", (long)(last + 1) );

	cw_d1_d2_precomp( P, tau, f, Output, d1,
		d2, last, scratch, drop, suff_lengths );
	/* tau is gamma_r, f is f_r, d1, d2 are correct */

	cw_bm_precomp( Shift, d1, d2, Char, suff_lengths, last );
	/* Shift contains k_bm */

	/* assume that the following is instrumented */
	cw_idx( Sprime, Output, tau_r, O, Shift, m_P );
}

static void CWN( void ) {
	/* assume d1, d2, d3_bar, suff_lengths correct */
	/* assume last is last_suff */
	cw_norm_precomp( Shift, d1, d2, d3_bar, last, suff_lengths );
	/* Shift is k_norm */

	/* assume that the following is instrumented */
	cw_idx( Sprime, Output, tau_r, O, Shift, m_P );
}



int main( int argc, char *argv[] ) {
	if( argc != 3 ) {
		fputs( "Usage: bench stringfile setfile f1 f2 f3 f4\n", stderr );
		exit( -1 );
	}
	if( ((stringfile = fopen( argv[1], "r" )) == NULL) ||
			((setfile = fopen( argv[2], "r" )) == NULL) /* ||
			((acfailfile = fopen( argv[3], "w" )) == NULL) ||
			((acoptfile = fopen( argv[4], "w" )) == NULL) ||
			((cwbmfile = fopen( argv[5], "w" )) == NULL) ||
			((cwnormfile = fopen( argv[6], "w" )) == NULL)*/ ) {
		fputs( "File opening problem", stderr );
		exit( -1 );
	}

	printf( "String read %ld\n", (long)readstring() );
	inittimer();
	while( readset() ) { 
		touch();
		if( P->next == NULL ) {
			/* single keyword set */
			KMP();
		}
		ACF();
		ACO();
		CWBM();
		CWN();
	}
	putchar('\n');
}
